import os
import logging
import json
import re
import torch
from argparse import Namespace


class Tracker:

    def __init__(self, log=None, epoch=0, iteration=0):
        self.log = log or []
        self.epoch = epoch
        self.iteration = iteration
        self.writer = None

    def save(self, fname):
        with open(fname, 'wt') as f:
            json.dump({'log': self.log, 'epoch': self.epoch, 'iteration': self.iteration}, f, indent=2)

    @classmethod
    def load(cls, fname):
        with open(fname) as f:
            return cls(**json.load(f))

    def best(self, early_stop, reverse=True):
        sort = sorted(self.log, key=lambda m: m['dev'][early_stop], reverse=reverse)
        return sort[0]

    def is_best(self, metrics, early_stop, iteration, epoch, reverse=True):
        self.iteration = iteration
        self.epoch = epoch
        better = True
        if self.log:
            best = self.best(early_stop, reverse=reverse)
            better = metrics['dev'][early_stop] >= best['dev'][early_stop] if reverse else metrics['dev'][early_stop] <= best['dev'][early_stop]
        self.log.append(metrics.copy())
        return better

    @classmethod
    def save_data(cls, data, fname, verbose=True):
        logging.info('Saving file {}'.format(fname))
        with open(fname, 'wt') as f:
            json.dump(data, f)

    @classmethod
    def find_checkpoints(cls, root, early_stop):
        files = [f for f in os.listdir(root) if f.endswith('.pt')]
        score_re = re.compile(r'{}=([0-9.\-]+)'.format(early_stop))
        files = []
        for f in os.listdir(root):
            scores = score_re.findall(f)
            if f.endswith('.pt') and scores:
                files.append((os.path.join(root, f), float(scores[0].rstrip('.'))))
        files.sort(key=lambda tup: tup[-1], reverse=True)
        return files

    @classmethod
    def clean_old_checkpoints(cls, root, early_stop, keep=5):
        files = cls.find_checkpoints(root, early_stop)
        if len(files) > keep:
            for f, s in files[keep:]:
                os.remove(f)

    def save_checkpoint(self, dout, metrics, early_stop, module, optimizer, args, link_best=False):
        fcheckpoint = os.path.join(dout, 'checkpoint.epoch={},iter={},{}={}.pt'.format(self.epoch, self.iteration, early_stop, metrics['dev'][early_stop]))
        torch.save({
            'metrics': metrics,
            'module': module.state_dict(),
            'optimizer': optimizer.state_dict(),
            'epoch': self.epoch,
            'iteration': self.iteration,
            'args': args,
        }, fcheckpoint)
        if link_best:
            fbest = os.path.join(dout, 'checkpoint.best.pt')
            if os.path.islink(fbest):
                os.unlink(fbest)
            os.symlink(fcheckpoint, fbest)

    @classmethod
    def load_best_checkpoint(cls, root, early_stop):
        files = cls.find_checkpoints(root, early_stop)
        if not files:
            raise Exception('No checkpoints found in {} with early stopping metric {}'.format(root, early_stop))
        fbest = files[0][0]
        logging.info('Loading {}'.format(fbest))
        return torch.load(fbest)



import logging


def get_logger(name, fout=None, level=logging.INFO, file_level=None):
    """
    Returns a `logging.Logger` object that outputs to stdout and optionally to disk.
    """
    logger = logging.Logger(name)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch = logging.StreamHandler()
    ch.setLevel(level)
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    if fout:
        fh = logging.FileHandler(fout)
        fh.setLevel(file_level or level)
        fh.setFormatter(formatter)
        logger.addHandler(fh)
    return logger



from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter


def get_default_parser():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('module', help='name of module')
    parser.add_argument('--name', help='name of experiment', default='default')
    parser.add_argument('--module_root', help='where to select modules from', default='modules')
    parser.add_argument('--data_root', help='where to store dataset', default='data')
    parser.add_argument('--exp_root', help='where to store experiments', default='exp')
    parser.add_argument('--batch', help='examples per batch', type=int, default=64)
    parser.add_argument('--epoch', help='number of epochs to run', type=int, default=50)
    parser.add_argument('--demb', help='word embedding size', type=int, default=400)
    parser.add_argument('--debug', help='only load this many examples for debug. 0 means off.', type=int, default=0)
    parser.add_argument('--gpus', help='gpus to use', nargs='*', default=[])
    return parser



import torch
from torch import nn
from torch.nn import functional as F
from vnlp.nn.functional import mask_invalid_scores


def expand_sentinel(sentinel, seq):
    batch_size, _, dfeat = seq.size()
    return sentinel.unsqueeze(0).unsqueeze(0).expand(batch_size, 1, dfeat)


class Pointer(nn.Module):
    """
    dot product pointer.
    """

    def forward(self, seq, lens, cond):
        scores = seq.bmm(cond.unsqueeze(2)).squeeze(2)
        return mask_invalid_scores(scores, lens)


class Attention(Pointer):
    """
    attend over the sequences `seq` using the condition `cond`.
    """

    def __init__(self, dhid):
        super().__init__()
        self.sentinel = nn.Parameter(torch.Tensor(dhid))
        nn.init.uniform_(self.sentinel, -0.1, 0.1)

    @classmethod
    def mix(cls, seq, raw_scores):
        scores = F.softmax(raw_scores, dim=1)
        return scores.unsqueeze(2).expand_as(seq).mul(seq).sum(1)

    def forward(self, seq, lens, cond):
        seq = torch.cat([expand_sentinel(self.sentinel, seq), seq], dim=1)
        raw_scores = super().forward(seq, lens+1, cond)
        return self.mix(seq, raw_scores)


class SelfAttention(nn.Module):
    """
    scores each element of the sequence with a linear layer and uses the normalized scores to compute a context over the sequence.
    """

    def __init__(self, dhid, scorer=None):
        super().__init__()
        self.scorer = scorer or nn.Linear(dhid, 1)
        self.sentinel = nn.Parameter(torch.Tensor(dhid))
        nn.init.uniform_(self.sentinel, -0.1, 0.1)

    def forward(self, inp, lens):
        inp = torch.cat([expand_sentinel(self.sentinel, inp), inp], dim=1)
        batch_size, seq_len, d_feat = inp.size()
        scores = self.scorer(inp.contiguous().view(-1, d_feat)).view(batch_size, seq_len)
        raw_scores = mask_invalid_scores(scores, lens+1)
        scores = F.softmax(raw_scores, dim=1)
        context = scores.unsqueeze(2).expand_as(inp).mul(inp).sum(1)
        return context


class Coattention(nn.Module):
    """
    one layer of coattention.
    """

    def __init__(self, dhid):
        super().__init__()
        self.trans = nn.Linear(dhid, dhid)
        self.q_sentinel = nn.Parameter(torch.Tensor(dhid))
        self.d_sentinel = nn.Parameter(torch.Tensor(dhid))
        nn.init.uniform_(self.q_sentinel, -0.1, 0.1)
        nn.init.uniform_(self.d_sentinel, -0.1, 0.1)

    def forward(self, q, q_len, d, d_len):
        batch_size, _, dfeat = q.size()
        q = torch.cat([expand_sentinel(self.q_sentinel, q), q], dim=1)
        d = torch.cat([expand_sentinel(self.d_sentinel, d), d], dim=1)
        q = self.trans(q)
        a = q.bmm(d.transpose(1, 2))
        aq = F.softmax(mask_invalid_scores(a, q_len+1), dim=1)
        ad = F.softmax(mask_invalid_scores(a.transpose(1, 2), d_len+1), dim=1)
        sd = q.transpose(1, 2).bmm(aq)
        sq = d.transpose(1, 2).bmm(ad)
        cd = sq.bmm(aq)
        cd, sq, sd = cd.transpose(1, 2), sq.transpose(1, 2), sd.transpose(1, 2)
        return cd[:, 1:], sq[:, 1:], sd[:, 1:]



class AnnotationPipeline:

    def __init__(self, stages=None):
        self.stages = stages or []

    def __call__(self, element):
        ann = {'orig': element}
        for stage in self.stages:
            stage(ann)
        return ann



import subprocess
import json
import logging


def run(command, only_if=None, fail_on_error=False):
    if only_if is None or only_if:
        try:
            out = subprocess.check_output(command).decode().strip()
            return out
        except Exception as e:
            if fail_on_error:
                raise e
            else:
                logging.error(e)



import os
import logging
import torch


class BaseDataset:

    def __init__(self, splits=None):
        self.splits = splits or {}

    def debug(self, size=5):
        if size:
            for name, split in self.splits.items():
                self.splits[name] = split.__class__(split[:size])

    def __repr__(self):
        order = sorted(list(self.splits.keys()))
        return '{}({})'.format(self.__class__.__name__, ', '.join(['{}: {}'.format(s, repr(self.splits[s])) for s in order]))

    def save(self, fname):
        torch.save(self, fname)

    @classmethod
    def load(cls, fname):
        return torch.load(fname)

    @classmethod
    def download(cls):
        raise NotImplementedError()

    @classmethod
    def pull(cls, fname, cache=None, debug=0):
        cache = cache or os.path.dirname(fname)
        if not os.path.isfile(fname):
            cls.download(cache).save(fname)
        data = cls.load(fname)
        if debug:
            data.debug(debug)
        return data



__version__ = '0.0.1'
from vnlp.data.dataset import BaseDataset
from vnlp.data.datasplit import DataSplit
from vnlp.data.example import Example
from vnlp.data.pipeline import AnnotationPipeline
from vnlp.data.stage import Tokenize, Numericalize

from vnlp.exp.grapher import Grapher
from vnlp.exp.tracker import Tracker

from vnlp.model.model import Model
from vnlp.model.module import BaseModule



import revtok
from vocab import Vocab


class Stage:

    def __call__(self, ann):
        raise NotImplementedError()


class Tokenize(Stage):

    @classmethod
    def tokenize(cls, sent):
        return revtok.tokenize(sent)

    @classmethod
    def detokenize(cls, tokens):
        return revtok.detokenize(tokens)

    def __call__(self, ann):
        ann['tokens'] = self.tokenize(ann['orig'])


class Numericalize(Stage):

    def __init__(self, vocab=None, append_ends=True, train=True, lower=None):
        self.vocab = vocab or Vocab()
        self.append_ends = append_ends
        self.train = train
        assert lower in {'all', 'first', None}
        self.lower = lower

    def __call__(self, ann):
        tokens = []
        for i, t in enumerate(ann['tokens']):
            t = t.strip()
            if self.lower == 'first' and i == 0 and (len(t) == 1 or t.upper() != t):
                t = t.lower()
            elif self.lower == 'all':
                t = t.lower()
            tokens.append(t)
        if self.append_ends:
            tokens = ['<S>'] + tokens + ['</S>']
        ann['num'] = self.vocab.word2index(tokens, train=self.train)



from torch import nn


class BaseModule(nn.Module):

    def __init__(self, args):
        super().__init__()
        self.args = args

    def featurize(self, batch, device=None):
        raise NotImplementedError()

    def compute_loss(self, out, batch):
        raise NotImplementedError()

    def extract_preds(self, out, batch):
        raise NotImplementedError()

    def graph(self, grapher, metrics, iteration):
        grapher.add_metrics(metrics, iteration=iteration)
        grapher.add_parameters(self, iteration)



import torch
import os
import re
import logging
import pprint
import json
import numpy as np
from argparse import Namespace
from torch import nn
from torch.nn import functional as F
from collections import defaultdict
from importlib import import_module
from ..utils.logging import get_logger
from ..exp.tracker import Tracker
from ..exp.grapher import Grapher


class Model:
    """
    args has:
    - gpu
    - exp_root
    - module
    - name
    - epoch
    - early_stop
    """

    def __init__(self, args, module):
        self.args = args
        self.device = torch.device('cpu')
        self.to(args.gpus)
        self.module = module

    def to(self, gpus):
        if gpus:
            self.device = torch.device('cuda:{}'.format(gpus[0]))
            if len(gpus) == 1:
                self.module = self.module.to(self.device)
            else:
                self.module = nn.DataParallel(self.module, device_ids=device)

    @property
    def module_instance(self):
        return self.module.module if isinstance(self.module, nn.DataParallel) else self.module

    @property
    def dout(self):
        return os.path.join(self.args.exp_root, self.args.module, self.args.name)

    @classmethod
    def from_module(cls, module, args, *vargs, module_root='modules', **kwargs):
        M = import_module('{}.{}'.format(module_root, module)).Module
        return cls(args, M(args, *vargs, **kwargs))

    def run_train(self, train, dev, optimizer=None, tracker=None, plot_module=False, logging_level=logging.INFO, verbose=True):
        if not os.path.isdir(self.dout):
            os.makedirs(self.dout)
        logger = get_logger('trainer', fout=os.path.join(self.dout, 'train.log'), level=logging_level)
        logger.info('Starting train with args:\n{}'.format(pprint.pformat(self.args)))
        tracker = tracker or Tracker()
        grapher = Grapher(self.dout)

        tracker.save_data(vars(self.args), 'config.json')

        if optimizer is None:
            optimizer = torch.optim.Adam(self.module.parameters())
        iteration = tracker.iteration
        for epoch in range(tracker.epoch, self.args.epoch):
            logger.info('Starting epoch {}'.format(epoch))
            loss = 0
            self.module.train()
            for batch in train.batch(self.args.batch, shuffle=True, verbose=verbose):
                optimizer.zero_grad()
                feat = self.module_instance.featurize(batch, device=self.device)
                if not grapher.plotted_module and plot_module:
                    grapher.plot_module(self.module_instance, feat)
                out = self.module(*feat)
                batch_loss = self.module_instance.compute_loss(out, batch)
                batch_loss.backward()
                optimizer.step()
                loss += batch_loss.item()
                iteration += len(batch)
            train_preds = self.run_pred(train)
            dev_preds = self.run_pred(dev)
            train_metrics = train.compute_metrics(train_preds)
            train_metrics['loss'] = loss
            metrics = {
                'iteration': iteration,
                'epoch': epoch,
                'train': train_metrics,
                'dev': dev.compute_metrics(dev_preds),
            }

            self.module_instance.graph(grapher, metrics, iteration)

            if tracker.is_best(metrics, self.args.early_stop, iteration=iteration, epoch=epoch):
                logging.info('Found new best! Saving to {}'.format(self.dout))
                tracker.save_checkpoint(self.dout, metrics, self.args.early_stop, self.module_instance, optimizer, self.args, link_best=True)
                tracker.clean_old_checkpoints(self.dout, self.args.early_stop)
                tracker.save_data(train_preds, os.path.join(self.dout, 'train.preds.json'))
                tracker.save_data(dev_preds, os.path.join(self.dout, 'dev.preds.json'))
                metrics['best'] = metrics['dev']
            logger.info('\n' + pprint.pformat(metrics))
            tracker.save(os.path.join(self.dout, 'tracker.json'))
        state = tracker.load_best_checkpoint(self.dout, self.args.early_stop)
        self.module_instance.load_state_dict(state['module'])
        optimizer.load_state_dict(state['optimizer'])

    def run_pred(self, dev, verbose=False):
        self.module.eval()
        preds = []
        for batch in dev.batch(self.args.batch, shuffle=False, verbose=verbose):
            out = self.module(*self.module_instance.featurize(batch, device=self.device))
            preds += self.module_instance.extract_preds(out, batch)
        return preds



import os
from tensorboardX import SummaryWriter


class Grapher:

    def __init__(self, dout):
        self.writer = SummaryWriter(os.path.join(dout, 'tensorboard'))
        self.plotted_module = False

    def plot_module(self, module, args):
        if not isinstance(args, (list, tuple)):
            args = (args, )
        self.writer.add_graph(module, args)
        self.plotted_module = True

    def add_metrics(self, metrics, iteration):
        keys = sorted(list(set(list(metrics['train'].keys()) + list(metrics['dev'].keys()))))
        for k in keys:
            v = {}
            if k in metrics['train']:
                v['train'] = metrics['train'][k]
            if k in metrics['dev']:
                v['dev'] = metrics['dev'][k]
            if v:
                self.writer.add_scalars(k, v, iteration)

    def add_embeddings(self, embeddings, labels, iteration):
        self.writer.add_embedding(embeddings, metadata=labels, iteration=iteration)

    def add_parameters(self, module, iteration):
        for name, p in module.named_parameters():
            self.writer.add_histogram(name, p.detach().numpy(), iteration)



import torch
import numpy as np


def pad(seqs, pad_value=0, device=None):
    """
    pads a list of variable-lengthed lists and returns a padded tensor and a lengths tensor.
    """
    lens = [len(s) for s in seqs]
    max_len = max(lens)
    return torch.tensor([s + [pad_value] * (max_len-l) for s, l in zip(seqs, lens)], device=device), torch.tensor(lens, device=device)


def mask(lens, invert=False, device=None, total_length=None):
    """
    return a boolean mask of valid positions corresponding to the given lengths.
    """
    max_len = total_length if total_length else lens.max().item()
    valid = False if invert else True
    return torch.tensor([[valid] * l + [not valid] * (max_len-l) for l in lens.tolist()], device=device)


def mask_invalid_scores(scores, lens):
    """
    returns a differentiable version of scores whose invalid positions are masked with `-np.inf`.
    """
    valid = mask(lens, device=scores.device, total_length=scores.size(1))
    while scores.dim() > valid.dim():
        valid = valid.unsqueeze(valid.dim()).expand_as(scores)
    neg = torch.zeros_like(valid, device=scores.device).float()
    neg.masked_fill_(1-valid, -np.inf)
    return scores * valid.float() + neg



import numpy as np
from tqdm import tqdm


class DataSplit(list):

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, len(self))

    def batch(self, batch_size, shuffle=False, verbose=True):
        copy = self[:]
        if shuffle:
            np.random.shuffle(copy)
        iterator = range(0, len(copy), batch_size)
        iterator = tqdm(iterator) if verbose else iterator
        for i in iterator:
            yield copy[i:i+batch_size]


class AnnotatedDataSplit(DataSplit):
    pass



# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

from setuptools import setup, find_packages, find_namespace_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))
frequirement = path.join(here, 'requirements.txt')


def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]


# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='vnlp',
    version='1.0.0',

    python_requires='>3.7',

    description='library for NLP and ML.',
    long_description=long_description,
    long_description_content_type='text/markdown',

    # The project's main homepage.
    url='https://gitlab.com/vzhong/vnlp',

    # Author details
    author='Victor Zhong',
    author_email='victor@victorzhong.com',

    # Choose your license
    license='Apache 2',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
    ],

    # What does your project relate to?
    keywords='text nlp machine-learning',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=parse_requirements(frequirement),

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['check-manifest', 'pdoc', 'twine'],
        'test': ['coverage', 'nose'],
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    include_package_data=True,

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files=[],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
    },
)



import torch
from torch import nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence


class PackedRNN(nn.Module):

    def __init__(self, rnn):
        """A wrapper class that packs input sequences and unpacks output sequences"""
        super().__init__()
        self.rnn = rnn

    @property
    def batch_first(self):
        return self.rnn.batch_first

    def forward(self, inputs, lengths, prev_state=None, total_length=None, pad_value=0):
        sorted_lens, sorted_indices = torch.sort(lengths, dim=0, descending=True)
        inputs = inputs[sorted_indices] if self.batch_first else inputs[:, sorted_indices] 
        packed_inputs = pack_padded_sequence(inputs, lengths=sorted_lens, batch_first=self.batch_first)
        packed_outputs, state = self.rnn(packed_inputs, prev_state)
        outputs, _ = pad_packed_sequence(packed_outputs, batch_first=self.batch_first, padding_value=pad_value, total_length=total_length or sorted_lens.max().item())
        _, unsorted_indices = torch.sort(sorted_indices, 0)
        outputs = outputs[unsorted_indices] if self.batch_first else outputs[:, unsorted_indices]
        if isinstance(state, (tuple, list)):
            state = [s[:, unsorted_indices] for s in state]
        else:
            state = state[:, unsorted_indices]
        return outputs, state



import os
import logging
import requests


def download(url, fname, skip_if_exists=True, chunk_size=1024):
    """
    Downloads a file from an URL.
    """
    if os.path.isfile(fname) and skip_if_exists:
        return
    logging.info('Downloading from {} to {}'.format(url, fname))
    r = requests.get(url, stream=True)
    with open(fname, 'wb') as f:
        for chunk in r.iter_content(chunk_size=chunk_size): 
            if chunk:
                f.write(chunk)
